/* FILE: evoc.h                    -*-Mode: c++-*-
 *
 * Stuff that should be moved eventually into the oc extension.
 * 
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 1999/05/19 15:30:46 $
 * Last modified by: $Author: dgp $
 */

#ifndef _NB_EVOC
#define _NB_EVOC

#include "oc.h"

#include <math.h>
#include <time.h>

/* End includes */     /* Optional directive to build.tcl */

#ifndef PI
#define PI             3.14159265358979323846
#endif

#ifndef SQRT2
#define SQRT2          1.41421356237309504880
#endif

#define SCRATCH_BUF_SIZE 1024  // Size of scratch space buffers
#define ASCIIPTR_BUF_SIZE (POINTERASCIIWIDTH+16)
/// Room + extra for %p representation of pointers


extern int Verbosity;
#ifndef DEBUGLVL
#define DEBUGLVL 0  // 0 => No debug code, 100 => Full debug code
#endif 

// Several useful macros
#define OC_SQ(x) ((x)*(x))
#define OC_MIN(x,y) ((x)<(y)?(x):(y))
#define OC_MAX(x,y) ((x)>(y)?(x):(y))
#define OC_ROUND(x) ((x)<0 ? ceil((x)-0.5) : floor((x)+0.5))
/// Note: rint() is not in the ANSI C spec, and in any event
///  may vary depending on "prevalent rounding mode".


//////////////////////////////////////////////////////////////////////////
// Class documentation structure
struct ClassDoc {
public:
  const CHAR *classname;
  const CHAR *maintainer;
  const CHAR *revision;
  const CHAR *revdate;
  ClassDoc(const CHAR *new_classname,const CHAR *new_maintainer,
	   const CHAR *new_revision,const CHAR *new_revdate);
};

//////////////////////////////////////////////////////////////////////////
// Tcl/Tk Interface

typedef struct {
  int major;        // Tcl major version
  int minor;        //     minor version
  int level;        //     release level
  int serial;       //     release serial
} OmfTclPatchLevelStruct;

// Patch level at run time (which may differ from compile time pl.)
extern OmfTclPatchLevelStruct Omf_RunningTclPatchLevel;

void Omf_SetPatchLevel(Tcl_Interp *interp,OmfTclPatchLevelStruct &tpl);
int Omf_ComparePatchLevel(const OmfTclPatchLevelStruct &tpl,
			  int major,int minor,int level,int serial);
/// Returns 0 on equal, -1 if tpl<major.minor.level.serial,
///                  or +1 if tpl>major.minor.level.serial

// C-pointer <--> ascii string conversion class.
//   Currently this routine maps things through %x or %lx, but this
// may change in the future to use a hash table.  The client should
// first call GetAsciiSize to determine how large a buffer is needed
// to hold the ascii-fied pointer string, then call PtrToAscii to
// get a string representation of the ptr.  This can be passed across
// to Tcl procs as needed.  Later, when a C routine is called by a
// Tcl proc having such a string, this string can be converted back
// to a valid ptr via AsciiToPtr.  NOTE: AsciiToPtr should ONLY be
// called with strings produced by PtrToAscii.  There is currently
// no check on this, but there will be if this class gets modified
// to use a hash table lookup scheme.
//
class Omf_AsciiPtr
{
private:
  static const ClassDoc class_doc;
  static const size_t ascii_string_width;
public:
  static size_t GetAsciiSize() { return ascii_string_width; }

  static void PtrToAscii(const void* ptr,char* buf);
  /// It is the responsibility of the calling routine to insure
  /// that buf is dimensioned to at least GetStringSize() bytes.

  static void* AsciiToPtr(const char* buf);
  /// NOTE 1: ONLY strings produced by Omf_AsciiPtr::PtrToAscii
  ///  should ever be sent to this routine. 
  /// Note 2: For consistency, I would prefer to make this function
  ///    static void AsciiToPtr(const char* buf,void * &ptr)
  ///  but gcc 2.7.2.1 (at least) screws up implicit casting
  ///  to void *&, apparently producing a temporary which never
  ///  gets written back into ptr.  So the above assignment seems
  ///  a safer option.  (The alternative is to be _real_ sure
  ///  that the client routine does an explicit cast to void *&.)
};

//////////////////////////////////////////////////////////////////////////
// Utility routines

// Random number generator
double UnifRand(unsigned int seed=0);
/// Returns random value in [0,1] with unif. distrib.
/// seed>0 => reinitialize generator with "seed"
/// Routines needing random numbers should call this rather than
/// the system calls rand() or random(), which may be system dependent.

// Routine to kill zombies under unix.  This is a NOP under windows
Tcl_CmdProc CullZombies;

//////////////////////////////////////////////////////////////////////////
// Byte reversal routines, for conversion between big-endian
// and little-endian representations
inline void flip2(void* input);
inline void flip4(void* input);
inline void flip8(void* input);  // May be too big to inline
void swap_bytes(void *buf,INT4m buflen);
void swap_words4(void *buf,INT4m count);
void swap_words8(void *buf,INT4m count);
void swap_buf(void *buf,INT4m wordsize,INT4m wordcount);
  // Converts the array 'buf' from big endian to little endian
  // (or vice versa), where each word is of width 'wordsize',
  // and there are 'wordcount' elements (words) in the array.

// inline codes...
inline void flip2(void* input)
{ // In place byte reversal for all 2 byte wide types.
  UCHAR temp;
  UCHAR* byte=(UCHAR*)input;
  temp=byte[0]; byte[0]=byte[1]; byte[1]=temp;
}

inline void flip4(void* input)
{ // In place byte reversal for all 4 byte wide types.
  UCHAR temp;
  UCHAR* byte=(UCHAR*)input;
  temp=byte[1]; byte[1]=byte[2]; byte[2]=temp;
  temp=byte[0]; byte[0]=byte[3]; byte[3]=temp;
}

inline void flip8(void* input)
{ // In place byte reversal for all 4 byte wide types.
  // Probably too big to be inlined.
  UCHAR temp;
  UCHAR* byte=(UCHAR*)input;
  temp=byte[3]; byte[3]=byte[4]; byte[4]=temp;
  temp=byte[2]; byte[2]=byte[5]; byte[5]=temp;
  temp=byte[1]; byte[1]=byte[6]; byte[6]=temp;
  temp=byte[0]; byte[0]=byte[7]; byte[7]=temp;
}

//////////////////////////////////////////////////////////////////////////
// Oc_Times function
//   This is a portable replacement for the Unix times(2) function
class Oc_TimeVal; // Forward declaration
void Oc_Times(Oc_TimeVal& cpu_time,Oc_TimeVal& wall_time);

//////////////////////////////////////////////////////////////////////////
// Oc_TimeVal class
//  NOTE: This class supports only non-negative times.
class Oc_TimeVal {
  friend void Oc_Times(Oc_TimeVal& cpu_time,Oc_TimeVal& wall_time);
private:
  static const ClassDoc class_doc;
  unsigned long ticks_per_second;
  unsigned long max_ticks;  // Largest allowed value for ticks
  /// The total number of ticks is overflow*(max_ticks+1)+ticks.
  unsigned long ticks;
  unsigned long overflow;
  void Print(FILE* fptr);  // For debugging
public:
  void Reset();
  void Reset(unsigned long _ticks_per_second,unsigned long _max_ticks);

  Oc_TimeVal();
  Oc_TimeVal(const Oc_TimeVal& time);
  Oc_TimeVal(unsigned long _ticks_per_second,unsigned long _max_ticks);

  unsigned long GetTicksPerSecond() const { return ticks_per_second; }
  unsigned long GetMaxTicks() const { return max_ticks; }
  unsigned long GetTicks() const { return ticks; }
  void SetTicks(unsigned long _ticks);
  unsigned long GetOverflow() const { return overflow; }
  void SetOverflow(unsigned long _overflow) { overflow=_overflow; }
  BOOL IsValid() const;
  operator double() const; // Returns time in seconds, in floating point
  Oc_TimeVal& operator=(const Oc_TimeVal& time);
  Oc_TimeVal& operator+=(const Oc_TimeVal& time);
  Oc_TimeVal& operator-=(const Oc_TimeVal& time);
  friend BOOL AreCompatible(const Oc_TimeVal& time1,
			    const Oc_TimeVal& time2);
  friend Oc_TimeVal operator+(const Oc_TimeVal& time1,
			      const Oc_TimeVal& time2);
  friend Oc_TimeVal operator-(const Oc_TimeVal& time1,
			      const Oc_TimeVal& time2);
};

#endif // _NB_EVOC
